package drawingclient;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.rmi.RemoteException;
import java.util.ArrayList;

import drawing.Curve;
import drawing.NoMoreLeftException;
import drawing.Queue;
import drawingserver.DrawingServer;

/**
 * Przechowywanie krzywych i buforowanie wysyania/odbierania (w celu odseparowania
 * ewentualnych opnie zwizanych z sieci od interfejsu).
 */

public class CurveBank {
	/**
	 * Lista zawierajca krzywe (obiekty klasy Curve).
	 */
	private ArrayList curves = new ArrayList();

	/**
	 * Klucz sesji. CurveBank podaje go przy ciganiu krzywych z serwera (eby serwer
	 * wiedzia co wysa).
	 */
	private String clientID;

	/**
	 * Referencja na interfejs serwera DrawingDrawing.
	 */
	private DrawingServer drawingServer;

	/**
	 * Czas co jaki klient prbuje cign nowe krzywe z serwera.
	 */
	private int refreshInterval = 100;

	/**
	 * Koleja cignitych krzywych, ktre nie zostay jeszcze narysowane.
	 */
	private Queue downloadedQueue = new Queue();

	/**
	 * Kolejka krzywych przeznaczonych do wysania.
	 */
	private Queue uploadQueue = new Queue();	

	/**
	 * Konstruktor. Nic nie robi.
	 */
	public CurveBank() {
	}

	/**
	 * Usu wszystkie krzywe z banku.
	 *
	 */
	public synchronized void dropAllCurves(){		
			curves = new ArrayList();
			uploadQueue = new Queue();		
			downloadedQueue = new Queue();				
	}
	
	/**
	 * Zapisz bank do strumienia. Suy do zapisywania obrazka (czyli listy krzywych) na dysk.
	 * Korzysta z mechanizmu serializacji.
	 * @param out Strumie do ktrego ma by zapisany obrazek
	 * @throws IOException
	 */
	public void saveToStream(OutputStream out) throws IOException{
		//ByteArrayOutputStream memoryOutputStream = new ByteArrayOutputStream();
        ObjectOutputStream serializer = new ObjectOutputStream(out);            
        serializer.writeObject(curves);            
        serializer.flush();
	}
	
	/**
	 * Wczytywanie banku z dysku.
	 * @param in Strumie, z ktrego zostanie wczytany obrazek
	 * @throws IOException
	 */
	public void loadFromStream(InputStream in) throws IOException{				
		ObjectInputStream deserializer = new ObjectInputStream(in);            
        try {
			ArrayList newCurves = (ArrayList)deserializer.readObject();
			
			for(int i=0;i<newCurves.size();i++){
				Curve c = (Curve)newCurves.get(i);				
				addCurve(c);
			}
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}             
	}
	
	/**
	 * @return Krzywa, ktra zostaa cignita z serwera ale jeszcze nie zostaa narysowana. 
	 * @throws NoMoreLeftException
	 */
	public Curve getNextCurveFromDownloadedQueue() throws NoMoreLeftException{
		return (Curve)downloadedQueue.poll();
	}

	/**
	 * Inicjalizacja banku krzywych, tzn. poczenie z serwerem i zapisanie unikalego identyfikatora sesji.
	 * @param drawingServer Wskanik na serwer (wczenij musiao nastpi poczenie RMI)
	 * @throws RemoteException
	 */
	public void init(DrawingServer drawingServer) throws RemoteException{
		this.drawingServer = drawingServer;
		
			clientID = drawingServer.register();
			System.out.println("Client registered as " + clientID);
			new BufferedDownloader();
			new BufferedUploader();					
	}
	
	/**
	 * Wylij krzyw na serwer, ale nie zapisuj jej lokalnie.
	 * @param c Krzywa do wysania
	 */
	public void sendCurveOnly(Curve c){
		c.setAuthor(clientID);
		uploadQueue.add(c);
	}

	/**
	 * Dodaj krzyw. Zostaje zapisana lokalnie i wysana na serwer.
	 * @param c Krzywa do wysania.
	 */
	public synchronized void addCurve(Curve c) {		
		c.setAuthor(clientID);
		
		curves.add(c);						
		uploadQueue.add(c);		
	}
	
	/**
	 * Listener do suchania, czy nie nastpi bd przy komunikacji z serwerem.
	 * Wyjtek zwizany z poczeniem
	 */
	/*ActionListener connectionErrorListener;
	public void addConnectionErrorListener(ActionListener connectionErrorListener){
		this.connectionErrorListener=connectionErrorListener;
		//new ActionEvent();
	}*/
	/**
	 * Referencja na komponent do ktrego s oddelegowywane bdy (poniewa
	 * CurveBank nie jest graficznym komponentem i nie moe wywietli okienka z bdem).
	 */
	private ErrorShowing errorShower;
	/** 
	 * Referencja na obiekt do ktrego maj by oddelegowywane bdy (JPanel_DrawingDrawing go implementuje)
	 * @param ErrorShower
	 */
	public void addErrorShower(ErrorShowing errorShower){
		this.errorShower=errorShower;
	}

	/**
	 * Wtek cigajcy krzywe z serwera.
	 * @author marek
	 *
	 */
	private class BufferedDownloader extends Thread {
		boolean stop = false;
		public BufferedDownloader(){
			start();
		}
		
		public void kill(){
			stop = true;
		}
		
		
		
		public void run() {

			try {

				while (true) {

					try {
						while (true) {
							if(stop) return;
							Curve downloadedCurve = drawingServer.downloadCurve(clientID);																											
							downloadedQueue.add(downloadedCurve);
							//System.out.println("Downloaded");
							curves.add(downloadedCurve);							
						}
					} catch (NoMoreLeftException e) {
					}

					sleep(refreshInterval);

				}

			} catch (Exception e) {
				// TODO Auto-generated catch block				
				errorShower.showError(e);
			}
		}
	}

	/**
	 * Wtek wysyajacy narysowane krzywe.
	 * @author marek
	 *
	 */
	private class BufferedUploader extends Thread {
		boolean stop = false;
		public BufferedUploader(){
			start();
		}
		
		public void kill(){
			stop=true;
		}
		
		public void run() {

			try {

				while (true) {

					try {
						while (true) {
							if(stop) return;
							drawingServer.uploadCurve(((Curve) uploadQueue
									.poll()));
						}
					} catch (NoMoreLeftException e) {
					}

					sleep(refreshInterval);

				}

			} catch (Exception e) {
				// TODO Auto-generated catch block
				errorShower.showError(e);				
			}
		}
	}

	/**
	 * 
	 * @return Lista wszystkich krzywych
	 */
	public ArrayList getCurves() {
		return curves;
	}
}
