package drawingclient;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;

import drawing.Curve;
import drawing.NoMoreLeftException;
import drawing.Pen;

/**
 * JPanel na ktrym powstaje rysunek. Rysuje obiekty klasy Curve. Uytkownik rysujc na nim
 * tworzy obiekty klasy Curve. 
 */

public class JPanel_Drawer extends javax.swing.JPanel implements PropertyChangeListener{

	/**
	 * Bank krzywych.
	 */
	private CurveBank curveBank;

	/**
	 * Czas co jaki panel uaktualnia obrazek o krzywe innych autorw. Krzywe
	 * narysowane lokalnie trafiaj na ekran natychmiast.
	 */
	private int refreshInterval = 50;
	
	/**
	 * Co ktra pozycja myszki ma by wykorzystywana przy tworzeniu krzywej. Mona w ten
	 * sposb ograniczy rozmiar krzywych (i zarazem danych przesyanych przez sie) ale
	 * tym samych uczyni je bardziej kanciatymi.
	 */
	private int useEveryNthPoint = 1;
	
	/**
	 * Maksymalna dugo krzywej. Krzywa zostanie przerwana po osigniciu maksymalnej
	 * dugoci i przeznaczona do wysania.
	 */
	private int maxCurveLength = 24;

	/**
	 * Kontekst graficzny.
	 */
	private Graphics2D graphics;
	
	//private UIData uiData;
	private Color backgroundColor = new Color(255,255,255);	
	
	/**
	 * Flaga "nie odwierzaj".
	 */
	private boolean dontRedraw = false;
	/**
	 * Flaga "teraz uytkonik rysuje".
	 */
	private boolean drawingNow = false;
	/**
	 * Flaga "obrazek wygadzany"
	 */
	private boolean antialiased = true;
	
	/**
	 * Flaga wyczy obrazek.
	 */
	private boolean clear = false;
	
	/**
	 * Aktualnie uywany pisak.
	 */
	private Pen actualPen = new Pen();

	/**
	 * Konstruktor.
	 * Inicjalizuje listenery uywane do rysowania (MouseMotionListener i MouseListener).
	 */
	public JPanel_Drawer() {
		super();
		//initGUI();

		MouseInput mouse = new MouseInput();
		addMouseListener(mouse);
		addMouseMotionListener(mouse);

		new Refresher();		
	}
	
	
	/**
	 * Zapisz krzyw w banku.
	 * @param c Krzywa do zapisania w banku krzywych (zostanie wysana za porednictwem
	 * serwera do pozostaych klientw)
	 */
	private void saveCurve(Curve c) {
		curveBank.addCurve(c);
	}

	/**
	 * Nanie krzyw na ekran.
	 * @param g Kontekst graficzny
	 * @param c Krzywa do narysowania
	 */
	private void drawCurve(Graphics g, Curve c) {
		
		//instruction parsing
		try {
			if(c.getInstruction().equals("clear")){
				curveBank.dropAllCurves();
				System.out.println("Got clear");
				repaint();
				return;
			}
		} catch (NullPointerException e) {}
		
		try {
			Point p1 = c.getNextPoint();
			while (true) {
				Point p2 = c.getNextPoint();
				usePen((Graphics2D)g, c.getPen());				
				g.drawLine(p1.x, p1.y, p2.x, p2.y);
				p1 = p2;
			}
		} catch (NoMoreLeftException e) {
		}
	}
	
	/**
	 * Ustaw pisak.
	 * @param pen Pisak ktrm bdzie si rysowa kad kolejn krzyw.
	 */
	public void setPen(Pen pen){
		actualPen = pen;
	}
	
	/**
	 * Uzywaj wskazanego pisaka do rysowania.
	 * @param g Kontekst graficzny
	 * @param pen Pisak
	 */
	private void usePen(Graphics2D g, Pen pen){			
		g.setColor(pen.getColor());
		g.setStroke(new BasicStroke(pen.getLineWidth()));
		if(antialiased){
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
					RenderingHints.VALUE_ANTIALIAS_ON);
		}
		
	}	
	
	/**
	 * Przeciona metoda odpowiedzialna za rysowanie. Odrysowywuje cay obrazek, wywoywana
	 * midzy innymi wtedy, kiedy panel zostanie przeskalowany. 
	 */
	public void paint(Graphics g){	
		if(dontRedraw)
			return;
		super.paint(g);
		setBackground(backgroundColor);
		if(antialiased)
			((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
					RenderingHints.VALUE_ANTIALIAS_ON);
				
		try {
			ArrayList curves = curveBank.getCurves();
			for(int i=0;i<curves.size();i++){
				drawCurve(g, ((Curve)curves.get(i)));
			}
		} catch (NullPointerException e) {}
		
		//System.out.print("ASD ");
	}

	/**
	 * Klasa odpowiedzialna za odczytywanie ruchw myszki. Tworzy krzywe.
	 *
	 */
	private class MouseInput extends MouseAdapter implements MouseMotionListener {
		Point lastPoint;

		Curve actualCurve;

		int excludeCounter;

		public MouseInput() {
		}
				
		public void mousePressed(MouseEvent e) {
			// TODO Auto-generated method stub
			drawingNow = true;
			excludeCounter = 0;
			//actualPen = getPen();
			graphics = (Graphics2D)getGraphics();
			if(antialiased)
				(graphics).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
						RenderingHints.VALUE_ANTIALIAS_ON);
			
			lastPoint = e.getPoint();

			actualCurve = new Curve();
			actualCurve.setPen(actualPen);
			//setPen(graphics, getPen());
			actualCurve.addPoint(lastPoint);
		}

		public void mouseReleased(MouseEvent e) {
			// TODO Auto-generated method stub
			actualCurve.addPoint(e.getPoint());
			saveCurve(actualCurve);			
			drawingNow = false;
		}

		public void mouseDragged(MouseEvent e) {
			if((excludeCounter++)%useEveryNthPoint!=0) return;			
			
			usePen(graphics, actualPen);
			graphics.drawLine(lastPoint.x, lastPoint.y, e.getPoint().x, e
					.getPoint().y);

			lastPoint = e.getPoint();
			actualCurve.addPoint(lastPoint);

			if (actualCurve.getPointCount() > maxCurveLength) {
				saveCurve(actualCurve);
				actualCurve = new Curve();
				actualCurve.setPen(actualPen);
				usePen(graphics, actualPen);
				actualCurve.addPoint(lastPoint);
			}
		}

		public void mouseMoved(MouseEvent e) {		
		}
	}	

	/**
	 * Ustaw bank krzywych, z ktrego bdzie korzysta panel rysujcy.
	 * @param curveBank Bank krzywych
	 */
	public void setCurveBank(CurveBank curveBank) {
		this.curveBank = curveBank;
	}

	/**
	 * Wtek odwierzajcy obrazek o krzywe nie narysowane lokalnie (czyli narysowane przez kogo innego).
	 * @author marek
	 *
	 */
	private class Refresher extends Thread {
		int pos = 0;

		public Refresher() {
			start();
		}

		public void run() {
			try {

				while (true) {

					try {
						graphics = (Graphics2D)getGraphics();						

						//Curves
						try {
							while (true){
								Curve downloadedCurve = curveBank.getNextCurveFromDownloadedQueue(); 
								drawCurve(graphics, downloadedCurve);
							}
						} catch (NoMoreLeftException e) {
						}
						
						//Instructions
						if(clear){							
							Curve c = new Curve();
							c.setInstruction("clear");								
							curveBank.sendCurveOnly(c);														
							clear = false;
						}						
						

					} catch (NullPointerException e) {
					}

					sleep(refreshInterval);

				}

			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();

			}

		}
	}
	
	/**
	 * Wyczy kartk. 
	 *
	 */
	public void clearSheetOfPaper(){
		clear = true;
		curveBank.dropAllCurves();
		repaint();
	}	

	/**
	 * Zdarzenie zmiany pisaka.
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		if(evt.getPropertyName().equals("pen")){
			actualPen = (Pen)evt.getNewValue();
		}
	}

	/**
	 * 
	 * @return Prawda, jeli rysunek jest wygadzany.
	 */
	public boolean isAntialiased() {
		return antialiased;
	}

	/**
	 * 
	 * @param antialiased Prawda, jeli rysunek ma by wygadzany.
	 */
	public void setAntialiased(boolean antialiased) {
		this.antialiased = antialiased;
		repaint();
	}		

}
